/* $Id$

This file is part of libmspgl
Copyright © 2007 Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#include "arb_shader_objects.h"
#include "arb_vertex_shader.h"
#include "except.h"
#include "extension.h"
#include "program.h"
#include "shader.h"

using namespace std;

namespace Msp {
namespace GL {

Program::Program():
	del_shaders(false),
	linked(false)
{
	require_extension("GL_ARB_shader_objects");
	require_extension("GL_ARB_vertex_shader");

	id=glCreateProgramObjectARB();
}

Program::Program(const string &vert, const string &frag):
	del_shaders(true),
	linked(false)
{
	require_extension("GL_ARB_shader_objects");
	require_extension("GL_ARB_vertex_shader");

	id=glCreateProgramObjectARB();
	attach_shader(*new Shader(VERTEX_SHADER, vert));
	attach_shader(*new Shader(FRAGMENT_SHADER, frag));
	link();
}

Program::~Program()
{
	if(del_shaders)
	{
		for(list<Shader *>::iterator i=shaders.begin(); i!=shaders.end(); ++i)
			delete *i;
	}
	glDeleteObjectARB(id);
}

void Program::attach_shader(Shader &shader)
{
	if(find(shaders.begin(), shaders.end(), &shader)==shaders.end())
	{
		glAttachObjectARB(id, shader.get_id());
		shaders.push_back(&shader);
	}
}

void Program::detach_shader(Shader &shader)
{
	list<Shader *>::iterator i=remove(shaders.begin(), shaders.end(), &shader);
	if(i!=shaders.end())
	{
		shaders.erase(i, shaders.end());
		glDetachObjectARB(id, shader.get_id());
	}
}

void Program::set_del_shaders(bool ds)
{
	del_shaders=ds;
}

void Program::bind_attribute(uint index, const string &name)
{
	glBindAttribLocationARB(id, index, name.c_str());
}

void Program::link()
{
	for(list<Shader *>::iterator i=shaders.begin(); i!=shaders.end(); ++i)
		if(!(*i)->get_compiled())
			(*i)->compile();

	glLinkProgramARB(id);
	if(!(linked=get_param(GL_LINK_STATUS)))
		throw CompileError(get_info_log());
}

int Program::get_param(GLenum param) const
{
	int value;
	glGetObjectParameterivARB(id, param, &value);
	return value;
}

string Program::get_info_log() const
{
	sizei len=get_param(GL_INFO_LOG_LENGTH);
	char log[len+1];
	glGetInfoLogARB(id, len+1, reinterpret_cast<GLsizei *>(&len), log);
	return string(log, len);
}

void Program::bind()
{
	if(!linked)
		throw InvalidState("Program is not linked");

	glUseProgramObjectARB(id);
	cur_prog=this;
}

int Program::get_uniform_location(const string &n) const
{
	return glGetUniformLocationARB(id, n.c_str());
}

/*void Program::uniform(int i, int v)
{
	glUniform1iARB(i, v);
}

void Program::uniform(int i, float x)
{
	glUniform1fARB(i, x);
}

void Program::uniform(int i, float x, float y)
{
	glUniform2fARB(i, x, y);
}

void Program::uniform(int i, float x, float y, float z)
{
	glUniform3fARB(i, x, y, z);
}

void Program::uniform(int i, float x, float y, float z, float w)
{
	glUniform4fARB(i, x, y, z, w);
}

void Program::uniform4(int i, const float *v)
{
	glUniform4fvARB(i, 1, v);
}

void Program::uniform_matrix4(int i, const float *v)
{
	glUniformMatrix4fvARB(i, 1, false, v);
}*/

void Program::unbind()
{
	if(cur_prog)
	{
		glUseProgramObjectARB(0);
		cur_prog=0;
	}
}

void Program::maybe_bind()
{
	if(cur_prog!=this)
		bind();
}

Program *Program::cur_prog=0;


Program::Loader::Loader(Program &p):
	prog(p)
{
	prog.set_del_shaders(true);

	add("vertex_shader",   &Loader::vertex_shader);
	add("fragment_shader", &Loader::fragment_shader);
	add("attribute",       &Loader::attribute);
}

Program::Loader::~Loader()
{
	prog.link();
}

void Program::Loader::vertex_shader(const string &src)
{
	prog.attach_shader(*new Shader(VERTEX_SHADER, src));
}

void Program::Loader::fragment_shader(const string &src)
{
	prog.attach_shader(*new Shader(FRAGMENT_SHADER, src));
}

void Program::Loader::attribute(uint i, const string &n)
{
	prog.bind_attribute(i, n);
}

} // namespace GL
} // namespace Msp
